# Generated by Django 3.0.3 on 2020-02-23 10:09

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('image_labelling_tool', '0003_auto_20180911_0954'),
    ]

    operations = [
        migrations.AlterField(
            model_name='labels',
            name='last_modified_by',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='modified_labels', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='labels',
            name='locked_by',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='locked_labels', to=settings.AUTH_USER_MODEL),
        ),
    ]
